/* GStreamer sunxifbsink plugin
 * Copyright (C) 2013 Harm Hanemaaijer <fgenfb@yahoo.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Suite 500,
 * Boston, MA 02110-1335, USA.
 */
/**
 * SECTION:element-sunxifbsink
 *
 * The sunxifbsink element intends implements a hardware accelerated
 * video sink for the console framebuffer of Allwinner A1x/A20-based
 * devices. The basis of the implementation is the optimized fbdev
 * sink as implemented in the GstFbdevFramebufferSink class.
 *
 * <refsect2>
 * <title>Property settings,<title>
 * <para>
 * The plugin comes with variety of configurable properties regulating
 * the size and frames per second of the video output, and various 
 * options regulating the rendering method (including rendering directly
 * to video memory and page flipping).
 * </para>
 * </refsect2>
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v videotestsrc ! sunxifbsink >/dev/null
 * ]|
 * Output the video test signal to the framebuffer. The redirect to
 * null surpressed interference from console text mode.
 * |[
 * gst-launch -v videotestsrc ! sunxifbsink full-screen=true
 * ]|
 * Run videotstsrc at full-screen resolution
 * |[
 * gst-launch -v videotestsrc horizontal_speed=10 ! sunxifbsink \
 * full-screen=true buffer-pool=true graphics-mode=true
 * ]|
 * This command illustrates some of the plugin's optimization features
 * by rendering to video memory with vsync and page flipping in
 * console graphics mode. There should be no tearing with page flipping/
 * vsync enabled. You might have to use the fps property to reduce the frame
 * rate on slower systems.
 * |[
 * gst-launch playbin uri=[uri] video-sink="sunxifbsink full-screen=true"
 * ]|
 * Use playbin while passing options to sunxifbsink.
 * </refsect2>
 * <refsect2>
 * <title>Caveats</title>
 * <para>
 * The actual implementation of the Linux framebuffer API varies between
 * systems, and methods beyond the most basic operating mode may not work
 * correctly on some systems. This primarily applies to page flipping
 * and vsync. The API implementation may be slower than expected on certain
 * hardware due to, for example, extra hidden vsyncs being performed in the
 * pan function. The "pan-does-vsync" option may help in that case.
 * </para>
 * </refsect2>
 */

/*
 * The hardware addressing portion of the plugin was adapted from
 * xf86-video-sunxifb, which has the following copyright message;
 *
 * Copyright © 2013 Siarhei Siamashka <siarhei.siamashka@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdint.h>
#include <math.h>
#include <linux/kd.h>
#include <glib/gprintf.h>

#include <gst/gst.h>
#include <gst/video/gstvideosink.h>
#include <gst/video/video.h>
#include <gst/video/video-info.h>
#include "gstsunxifbsink.h"
#include <ion_mem_alloc.h>
#include "sunxi_tr.h"

GST_DEBUG_CATEGORY_STATIC (gst_sunxifbsink_debug_category);
#define GST_CAT_DEFAULT gst_sunxifbsink_debug_category

static inline void GST_SUNXIFBSINK_MESSAGE_OBJECT (GstSunxifbsink * sunxifbsink,
const gchar *message) {
  if (!sunxifbsink->fbdevframebuffersink.framebuffersink.silent)
    g_print ("%s.\n", message);
  else
    GST_INFO_OBJECT (sunxifbsink,"%"GST_PTR_FORMAT, message);
}

#define ALIGNMENT_GET_ALIGN_BYTES(offset, align) \
    (((align) + 1 - ((offset) & (align))) & (align))
#define ALIGNMENT_GET_ALIGNED(offset, align) \
    ((offset) + ALIGNMENT_GET_ALIGN_BYTES(offset, align))
#define ALIGNMENT_APPLY(offset, align) \
    offset = ALIGNMENT_GET_ALIGNED(offset, align);

/* Class function prototypes. */
static gboolean gst_sunxifbsink_open_hardware (
    GstFramebufferSink *framebuffersink, GstVideoInfo *info,
    gsize *video_memory_size, gsize *pannable_video_memory_size);
static void gst_sunxifbsink_close_hardware (
    GstFramebufferSink *framebuffersink);
static GstVideoFormat *gst_sunxifbsink_get_supported_overlay_formats (
    GstFramebufferSink *framebuffersink);
static gboolean gst_sunxifbsink_get_overlay_video_alignment (
    GstFramebufferSink *framebuffersink, GstVideoInfo *video_info,
    GstFramebufferSinkOverlayVideoAlignment *video_alignment,
    gint *overlay_align, gboolean *video_alignment_matches);
static gboolean gst_sunxifbsink_prepare_overlay (
    GstFramebufferSink *framebuffersink, GstVideoFormat format);
static GstFlowReturn gst_sunxifbsink_show_overlay (
    GstFramebufferSink *framebuffersink, GstMemory *memory);

static gboolean gst_sunxifbsink_reserve_layer (GstSunxifbsink *sunxifbsink);
static void gst_sunxifbsink_release_layer (GstSunxifbsink *sunxifbsink);
static gboolean gst_sunxifbsink_show_layer (GstSunxifbsink *sunxifbsink);
static void gst_sunxifbsink_hide_layer (GstSunxifbsink *sunxifbsink);

enum
{
  PROP_0,
};

#define GST_SUNXIFBSINK_TEMPLATE_CAPS \
        GST_VIDEO_CAPS_MAKE ("RGB") \
        "; " GST_VIDEO_CAPS_MAKE ("BGR") \
        "; " GST_VIDEO_CAPS_MAKE ("RGBx") \
        "; " GST_VIDEO_CAPS_MAKE ("BGRx") \
        "; " GST_VIDEO_CAPS_MAKE ("xRGB") \
        "; " GST_VIDEO_CAPS_MAKE ("xBGR") \
        "; " GST_VIDEO_CAPS_MAKE ("NV12") \
        "; " GST_VIDEO_CAPS_MAKE ("NV21") \
        "; " GST_VIDEO_CAPS_MAKE ("YV12") \
        "; " GST_VIDEO_CAPS_MAKE ("I420") \
        "; " GST_VIDEO_CAPS_MAKE ("YUY2") \
        "; " GST_VIDEO_CAPS_MAKE ("UYVY") \
        "; " GST_VIDEO_CAPS_MAKE ("Y444") \
        "; " GST_VIDEO_CAPS_MAKE ("AYUV") ", " \
        "framerate = (fraction) [ 0, MAX ], " \
        "width = (int) [ 1, MAX ], " "height = (int) [ 1, MAX ]"

static GstStaticPadTemplate gst_sunxifbsink_sink_template =
GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_SUNXIFBSINK_TEMPLATE_CAPS)
    );

static GstVideoFormat sunxifbsink_supported_overlay_formats_table[] = {
  /* List the formats that support odds widths first. */
  GST_VIDEO_FORMAT_YV12,
  GST_VIDEO_FORMAT_I420,
  GST_VIDEO_FORMAT_NV12,
  GST_VIDEO_FORMAT_NV21,
  GST_VIDEO_FORMAT_AYUV,
  GST_VIDEO_FORMAT_BGRx,
  /* These formats do not properly support odd widths. */
  GST_VIDEO_FORMAT_YUY2,
  GST_VIDEO_FORMAT_UYVY,
  GST_VIDEO_FORMAT_Y444,
  GST_VIDEO_FORMAT_UNKNOWN
};

#define TRANSFORM_DEV_TIMEOUT 200
#define ALIGN_32B(x) (((x) + (31)) & ~(31))
#define ALIGN_16B(x) (((x) + (15)) & ~(15))

/* Class initialization. */

#define gst_sunxifbsink_parent_class fbdevframebuffersink_parent_class
G_DEFINE_TYPE_WITH_CODE (GstSunxifbsink, gst_sunxifbsink,
  GST_TYPE_FBDEVFRAMEBUFFERSINK,
  GST_DEBUG_CATEGORY_INIT (gst_sunxifbsink_debug_category, "sunxifbsink", 0,
  "debug category for sunxifbsink element"));

static void
gst_sunxifbsink_class_init (GstSunxifbsinkClass* klass)
{
/*  GObjectClass *gobject_class = G_OBJECT_CLASS (klass); */
  GstFramebufferSinkClass *framebuffer_sink_class =
      GST_FRAMEBUFFERSINK_CLASS (klass);

/*  gobject_class->set_property = gst_sunxifbsink_set_property;
    gobject_class->get_property = gst_sunxifbsink_get_property; */

  /* Setting up pads and setting metadata should be moved to
     base_class_init if you intend to subclass this class. */
  gst_element_class_add_pad_template (GST_ELEMENT_CLASS(klass),
      gst_static_pad_template_get (&gst_sunxifbsink_sink_template));

  gst_element_class_set_static_metadata (GST_ELEMENT_CLASS(klass),
      "Accelerated console framebuffer video sink for sunxi-based devices",
      "Sink/Video",
      "sunxi framebuffer sink",
      "Harm Hanemaaijer <fgenfb@yahoo.com>");

  framebuffer_sink_class->open_hardware =
      GST_DEBUG_FUNCPTR (gst_sunxifbsink_open_hardware);
  framebuffer_sink_class->close_hardware =
      GST_DEBUG_FUNCPTR (gst_sunxifbsink_close_hardware);
  framebuffer_sink_class->get_supported_overlay_formats =
      GST_DEBUG_FUNCPTR (gst_sunxifbsink_get_supported_overlay_formats);
  framebuffer_sink_class->get_overlay_video_alignment =
      GST_DEBUG_FUNCPTR (gst_sunxifbsink_get_overlay_video_alignment);
  framebuffer_sink_class->prepare_overlay =
      GST_DEBUG_FUNCPTR (gst_sunxifbsink_prepare_overlay);
  framebuffer_sink_class->show_overlay =
      GST_DEBUG_FUNCPTR (gst_sunxifbsink_show_overlay);
}

/* Class member functions. */

static void
gst_sunxifbsink_init (GstSunxifbsink *sunxifbsink) {
	GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "-->sunxifbsink init");
}

static gboolean
gst_sunxifbsink_open_hardware (GstFramebufferSink *framebuffersink,
    GstVideoInfo *info, gsize *video_memory_size,
    gsize *pannable_video_memory_size)
{
  GstSunxifbsink *sunxifbsink = GST_SUNXIFBSINK (framebuffersink);
  GstFbdevFramebufferSink *fbdevframebuffersink =
      GST_FBDEVFRAMEBUFFERSINK (framebuffersink);
  unsigned long arg[4] = {0};
  
  if (!gst_fbdevframebuffersink_open_hardware(framebuffersink, info,
      video_memory_size, pannable_video_memory_size))
    return FALSE;

  sunxifbsink->hardware_overlay_available = FALSE;

  if (framebuffersink->use_hardware_overlay == FALSE)
    return TRUE;

  sunxifbsink->fd_disp = open ("/dev/disp", O_RDWR);

  if (sunxifbsink->fd_disp < 0)
    return TRUE;

  sunxifbsink->fd_transform = open("/dev/transform",O_RDWR);

  printf("fd_transform:%d\n",sunxifbsink->fd_transform);
  if (sunxifbsink->fd_transform < 0)
    return TRUE;

  arg[0] = (unsigned long)&(sunxifbsink->transform_channel);
  if(ioctl(sunxifbsink->fd_transform, TR_REQUEST,(void*)arg)<0){
	  printf("#######error: tr_request failed!");
	  return -1;
  }
  printf("request tr channel	= 0x%x",sunxifbsink->transform_channel);
  
  //* set tr timeout
  arg[0] = sunxifbsink->transform_channel;
  arg[1] = TRANSFORM_DEV_TIMEOUT;
  if(ioctl(sunxifbsink->fd_transform, TR_SET_TIMEOUT,(void*)arg) != 0)
  {
	  printf("#######error: tr_set_timeout failed!");
	  return -1;
  }

  
  /* Get the ID of the screen layer. */
 #if 0
  if (ioctl (fbdevframebuffersink->fd, sunxifbsink->framebuffer_id == 0 ?
  FBIOGET_LAYER_HDL_0 : FBIOGET_LAYER_HDL_1, &sunxifbsink->gfx_layer_id)) {
    close(sunxifbsink->fd_disp);
    return TRUE;
  }
 #endif
  
  if (!gst_sunxifbsink_reserve_layer(sunxifbsink)) {
    close(sunxifbsink->fd_disp);
    return TRUE;
  }

  sunxifbsink->layer_is_visible = FALSE;
  sunxifbsink->hardware_overlay_available = TRUE;
  GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "Hardware overlay available");

  return TRUE;
}

static void
gst_sunxifbsink_close_hardware (GstFramebufferSink *framebuffersink) {
  GstSunxifbsink *sunxifbsink = GST_SUNXIFBSINK (framebuffersink);
  GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "--> sunxifbsink close");
  struct SunxiMemOpsS* ops =  GetMemAdapterOpsS();

  if (sunxifbsink->hardware_overlay_available) {
    gst_sunxifbsink_hide_layer(sunxifbsink);
    gst_sunxifbsink_release_layer(sunxifbsink);
  }
  /* Before calling close_hardware, use_hardware_overlay is expected to have
     been reset to the original value it had when open_hardware was called. */
  if (framebuffersink->use_hardware_overlay)
    close(sunxifbsink->fd_disp);

  gst_fbdevframebuffersink_close_hardware (framebuffersink);
  if(sunxifbsink->rotate_addr_phy[0] != NULL)
  {
	SunxiMemPfree(ops,sunxifbsink->rotate_addr_phy[0]);
  }
  
  if(sunxifbsink->rotate_addr_phy[1] != NULL)
  {
	SunxiMemPfree(ops,sunxifbsink->rotate_addr_phy[1]);
  }

  if(sunxifbsink->fd_transform >= 0)
  	close(sunxifbsink->fd_transform);
}

static GstVideoFormat *
gst_sunxifbsink_get_supported_overlay_formats (
    GstFramebufferSink *framebuffersink)
{

  return sunxifbsink_supported_overlay_formats_table;
}

/* Return the video alignment (top/bottom/left/right padding and stride
   alignment for each plane) that  is required to display the overlay
   described by video_info. Also returns the alignment requirement of the
   start address of the overlay in video memory. video_alignment_matches
   is set to TRUE if the alignment defined by video_info did not have to be
   adjusted, FALSE otherwise. The function returns TRUE if hardware overlay
   with given video info is supported, FALSE otherwise. */

gboolean
gst_sunxifbsink_get_overlay_video_alignment(GstFramebufferSink *framebuffersink,
    GstVideoInfo *video_info, GstFramebufferSinkOverlayVideoAlignment *
    video_alignment, gint *overlay_align, gboolean *video_alignment_matches)
{
  GstVideoFormat format;
  format = GST_VIDEO_INFO_FORMAT (video_info);
  if (format == GST_VIDEO_FORMAT_I420 ||
      format == GST_VIDEO_FORMAT_YV12 ||
      format == GST_VIDEO_FORMAT_NV12 ||
      format == GST_VIDEO_FORMAT_NV21) {
    if (GST_VIDEO_INFO_WIDTH (video_info) & 1)
      /* Hardware overlay not supported for odd widths for all planar formats
         except Y444. Although it almost works for odd widths, there is an
         artifact line at the right of the scaled area, related to the
         alignment requirements of the width. */
      return FALSE;
  }
  /* When uses other formats, some artifacts have been observed when the width
     is odd, but for now leave support for odd widths enabled. */
  *overlay_align = 15;
  /* For the Allwinner hardware overlay, scanlines need to be aligned to pixel
     boundaries with a minimum alignment of word-aligned. This is a good match
     for the buffer format generally provided by upstream, so direct video
     memory buffer pool streaming is almost always possible. */
  gst_framebuffersink_set_overlay_video_alignment_from_scanline_alignment (
      framebuffersink, video_info, 3, TRUE, video_alignment,
      video_alignment_matches); 
  return TRUE;
}

/*
 * For the prepare overlay and show overlay functions, the parameters are
 * stored in the following fields:
 * framebuffersink->overlay_plane_offset[i] is the offset in bytes of each
 *     plane. Any top or left padding returned by get_overlay_video_alignment()
 *     will come first.
 * framebuffersink->overlay_scanline_offset[i] is the offset in bytes of the
 *     first pixel of each scanline for each plane (corresponding with the left
 *     padding * bytes per pixel). Usually 0.
 * framebuffersink->overlay_scanline_stride[i] is the scanline stride in bytes
 *     of each plane.
 * framebuffersink->videosink.width is the source width.
 * framebuffersink->videosink.height is the source height.
 * framebuffersink->video_rectangle.x is the destination x coordinate.
 * framebuffersink->video_rectangle.y is the destination y coordinate.
 * framebuffersink->video_rectangle.w is the destination width.
 * framebuffersink->video_rectangle.h is the destination height.
 */

static gboolean
gst_sunxifbsink_prepare_overlay (GstFramebufferSink *framebuffersink,
    GstVideoFormat format)
{
  GstSunxifbsink *sunxifbsink = GST_SUNXIFBSINK (framebuffersink);
  gchar s[256];
  
  g_sprintf(s,"---->sunxifb pre overlay(SCWxSCN=%d x %d,out_rec=[%d %d,%d,%d])",
  	framebuffersink->videosink.width , 
  	framebuffersink->videosink.height,
  	framebuffersink->video_rectangle.x,
  	framebuffersink->video_rectangle.y,
  	framebuffersink->video_rectangle.w,
  	framebuffersink->video_rectangle.h);
  GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, s);
  
  if (sunxifbsink->layer_is_visible)
    gst_sunxifbsink_hide_layer(sunxifbsink);

  sunxifbsink->overlay_format = format;

  return TRUE;
}

int hwRotateVideoPicture(GstSunxifbsink *sunxifbsink,tr_info *info)
{
    int ret = 0;
    unsigned long arg[4] = {0};
    tr_info tTrInfo;
	memcpy(&tTrInfo, info, sizeof(tr_info));
setup_tr:
    //* setup rotate
    arg[0] = sunxifbsink->transform_channel;
    arg[1] = (unsigned long)&tTrInfo;
    arg[2] = 0;
    arg[3] = 0;

    if(ioctl(sunxifbsink->fd_transform, TR_COMMIT,(void*)arg) != 0)
    {
        return -1;
    }

    //* wait
    arg[0] = sunxifbsink->transform_channel;
    arg[1] = 0;
    arg[2] = 0;
    arg[3] = 0;

    int nTimeOut = 0;
    ret = ioctl(sunxifbsink->fd_transform,TR_QUERY,(void*)arg);

    while(1)// 0 : suceef; 1:busy ; -1:timeOut
    {
        ret = ioctl(sunxifbsink->fd_transform, TR_QUERY,(void*)arg);
		if(ret == 1)
		{
        	usleep(1*1000);
		}
		else
			break;
    }

    //* if the tr is timeout ,we should setup tr again
    if(ret == -1)
    {
        goto setup_tr;
    }

    return ret;
}




static GstFlowReturn
gst_sunxifbsink_show_memory_yuv_planar (GstFramebufferSink *framebuffersink,
   	GstVideoFormat format,GstMemory *mem)
{
  GstSunxifbsink *sunxifbsink = GST_SUNXIFBSINK (framebuffersink);
  GstFbdevFramebufferSink *fbdevframebuffersink =
	  GST_FBDEVFRAMEBUFFERSINK (framebuffersink);
	struct disp_layer_config layer_config;
	unsigned long tmp[4];
	GstMapInfo mapinfo;
	char * phymem_start = 0;
	struct SunxiMemOpsS* ops =  GetMemAdapterOpsS();
	tr_info trans_info;
	static int m = 0;
	unsigned int width_align;
	unsigned int height_align;
	int rect_width;
	int rect_height;
	int rotate_enable = 0;
	tr_mode rt_mode;
	if(framebuffersink->rotate_angle_property > 0)
	{
		rotate_enable = 1;
		rt_mode = framebuffersink->rotate_angle_property;
	}
	else
	{
		rotate_enable = 0;
	}
	
	memset(&trans_info, 0, sizeof(tr_info));
	gst_memory_map(mem, &mapinfo, 0);

	memset(&layer_config, 0, sizeof(struct disp_layer_config));

	phymem_start = SunxiMemGetPhysicAddressCpu(ops, mapinfo.data);
	
	SunxiMemGetActualSize(ops,&rect_width,&rect_height);
	if (format == GST_VIDEO_FORMAT_Y444) {
	  layer_config.info.fb.addr[0] = (unsigned long long )phymem_start;
	  layer_config.info.fb.addr[1] = (unsigned long long )(phymem_start + framebuffersink->overlay_plane_offset[1]);
	  layer_config.info.fb.addr[2] = (unsigned long long )(phymem_start + framebuffersink->overlay_plane_offset[2]);
	  layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
		/ (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
		0, 8)
		* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
	  layer_config.info.fb.size[1].width = framebuffersink->overlay_scanline_stride[0]
		/ (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
		0, 8)
		* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
	  layer_config.info.fb.size[2].width = framebuffersink->overlay_scanline_stride[0]
		/ (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
		0, 8)
		* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
	  layer_config.info.fb.size[0].height = framebuffersink->videosink.height;
	  layer_config.info.fb.size[1].height = framebuffersink->videosink.height;
	  layer_config.info.fb.size[2].height = framebuffersink->videosink.height;
	  layer_config.info.fb.format = DISP_FORMAT_YUV444_P;
	  rotate_enable = 0;
	}
	else if (format == GST_VIDEO_FORMAT_NV12
		|| format == GST_VIDEO_FORMAT_NV21) {
	  if(format == GST_VIDEO_FORMAT_NV12)  
		layer_config.info.fb.format = DISP_FORMAT_YUV420_SP_UVUV;
	  else
		layer_config.info.fb.format = DISP_FORMAT_YUV420_SP_VUVU;
	  layer_config.info.fb.addr[0] = (unsigned long long )phymem_start;
	  layer_config.info.fb.addr[1] = (unsigned long long )(phymem_start + framebuffersink->overlay_plane_offset[1]);

	  layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
		/ (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
		0, 8)
		* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
	  layer_config.info.fb.size[1].width = (framebuffersink->overlay_scanline_stride[0]
		/ (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
		0, 8)
		* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8))/2;
	  
	  layer_config.info.fb.size[1].height = framebuffersink->videosink.height/2;
	  layer_config.info.fb.size[2].height = framebuffersink->videosink.height/2;

	  /*g_sprintf(s,"-->yuv planar(fb_off=0x%08X),start=0x%08x,p1_of=%x,p2_of=%x,
		str1=%d,vh=%d,w0=%d,(%d %d,%dx%d),%dx%d",framebuffer_offset, \
		fbdevframebuffersink->fixinfo.smem_start, \
		framebuffersink->overlay_plane_offset[1], \
		framebuffersink->overlay_plane_offset[2], \
		framebuffersink->overlay_scanline_stride[0], \
		framebuffersink->videosink.height, \
		layer_config.info.fb.size[0].width, \
		framebuffersink->video_rectangle.x, \
		framebuffersink->video_rectangle.y, \
		framebuffersink->video_rectangle.w, \
		framebuffersink->video_rectangle.h, \
		framebuffersink->videosink.width,framebuffersink->videosink.height);
	  GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, s);*/
	
	}
	else {
	  layer_config.info.fb.format = DISP_FORMAT_YUV420_P;	
	  layer_config.info.fb.addr[0] = (unsigned long long )phymem_start;
	  if (format == GST_VIDEO_FORMAT_I420) {
		layer_config.info.fb.addr[1] = (unsigned long long )(phymem_start + framebuffersink->overlay_plane_offset[1]);
		layer_config.info.fb.addr[2] = (unsigned long long )(phymem_start + framebuffersink->overlay_plane_offset[2]);
		layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
		/ (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
		0, 8)
		* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
		layer_config.info.fb.size[0].height = framebuffersink->videosink.height;
		layer_config.info.fb.size[1].width = framebuffersink->videosink.width/2;
		layer_config.info.fb.size[1].height = framebuffersink->videosink.height/2;
		layer_config.info.fb.size[2].width= framebuffersink->videosink.width/2;
		layer_config.info.fb.size[2].height = framebuffersink->videosink.height/2;
	  }
	  else {
		/* GST_VIDEO_FORMAT_YV12 */
		layer_config.info.fb.addr[1] = (unsigned long long )(phymem_start + framebuffersink->overlay_plane_offset[2]);
		layer_config.info.fb.addr[2] = (unsigned long long )(phymem_start + framebuffersink->overlay_plane_offset[1]);

		layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
		/ (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
		0, 8)* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
		layer_config.info.fb.size[0].height = framebuffersink->videosink.height;
		layer_config.info.fb.size[1].width = framebuffersink->videosink.width/2;
		layer_config.info.fb.size[1].height = framebuffersink->videosink.height/2;
		layer_config.info.fb.size[2].width= framebuffersink->videosink.width/2;
		layer_config.info.fb.size[2].height = framebuffersink->videosink.height/2;
	  }
	}
	
	if(rotate_enable)
	{
		width_align = ALIGN_32B(layer_config.info.fb.size[0].width);
		height_align = ALIGN_32B(layer_config.info.fb.size[0].height);
		if(sunxifbsink->rotate_addr_phy[0] == NULL && sunxifbsink->rotate_addr_phy[1] == NULL)
		{
			unsigned int buffer_len = width_align*height_align*3/2;
			sunxifbsink->rotate_addr_phy[0] = (char *)SunxiMemPalloc(ops,buffer_len);
			if(sunxifbsink->rotate_addr_phy[0] == NULL)
			{
				printf("no physical memory when rotate!\n");
				return GST_FLOW_ERROR;
			}
			sunxifbsink->rotate_addr_phy[1] = (char *)SunxiMemPalloc(ops,buffer_len);
			if(sunxifbsink->rotate_addr_phy[1] == NULL)
			{
				SunxiMemPfree(ops,sunxifbsink->rotate_addr_phy[0]);
				printf("no physical memory when rotate!\n");
				return GST_FLOW_ERROR;
				
			}
			memset(sunxifbsink->rotate_addr_phy[0],0,buffer_len);
			memset(sunxifbsink->rotate_addr_phy[1],0,buffer_len);
			SunxiMemFlushCache(ops,sunxifbsink->rotate_addr_phy[0],buffer_len);
			SunxiMemFlushCache(ops,sunxifbsink->rotate_addr_phy[1],buffer_len);
		}
	
		trans_info.mode = rt_mode;
		trans_info.src_frame.fmt = TR_FORMAT_YUV420_P;
		trans_info.src_frame.laddr[0] = (unsigned long)phymem_start;
		trans_info.src_frame.laddr[1] = (unsigned long)layer_config.info.fb.addr[1];
		trans_info.src_frame.laddr[2] = (unsigned long)layer_config.info.fb.addr[2];
	
		trans_info.src_frame.pitch[0] = ALIGN_32B(layer_config.info.fb.size[0].width);
		trans_info.src_frame.pitch[1] = ALIGN_32B(layer_config.info.fb.size[0].width)/2;
		trans_info.src_frame.pitch[2] = ALIGN_32B(layer_config.info.fb.size[0].width)/2;
		trans_info.src_frame.height[0] = ALIGN_32B(layer_config.info.fb.size[0].height);
		trans_info.src_frame.height[1] = ALIGN_32B(layer_config.info.fb.size[1].height)/2;
		trans_info.src_frame.height[2] = ALIGN_32B(layer_config.info.fb.size[2].height)/2;
	
		trans_info.src_rect.x = 0;
		trans_info.src_rect.y = 0;
		trans_info.src_rect.w = rect_width;
		trans_info.src_rect.h = rect_height;
		
		trans_info.dst_frame.fmt = TR_FORMAT_YUV420_P;		
		trans_info.dst_frame.laddr[0] = (unsigned long)SunxiMemGetPhysicAddressCpu(ops,(void *)sunxifbsink->rotate_addr_phy[(++m)%2]);
		trans_info.dst_frame.laddr[1] = trans_info.dst_frame.laddr[0] + width_align * height_align;
		trans_info.dst_frame.laddr[2] = trans_info.dst_frame.laddr[0] + width_align * height_align*5/4;
	
		if(trans_info.mode == TR_ROT_180)
		{
			trans_info.dst_frame.pitch[0] = width_align;
			trans_info.dst_frame.pitch[1] = width_align/2;
			trans_info.dst_frame.pitch[2] = width_align/2;
			
			trans_info.dst_frame.height[0] = height_align;
			trans_info.dst_frame.height[1] = height_align/2;
			trans_info.dst_frame.height[2] = height_align/2;
	
			trans_info.dst_rect.x = 0;
			trans_info.dst_rect.y = 0;
			trans_info.dst_rect.w = width_align;
			trans_info.dst_rect.h = height_align;
		}
		else
		{
			trans_info.dst_frame.pitch[0] = height_align;
			trans_info.dst_frame.pitch[1] = height_align/2;
			trans_info.dst_frame.pitch[2] = height_align/2;
			
			trans_info.dst_frame.height[0] = width_align;
			trans_info.dst_frame.height[1] = width_align/2;
			trans_info.dst_frame.height[2] = width_align/2;
			
			trans_info.dst_rect.x = 0;
			trans_info.dst_rect.y = 0;
			trans_info.dst_rect.w = height_align;
			trans_info.dst_rect.h = width_align;
		}
	
		hwRotateVideoPicture(sunxifbsink, &trans_info);

		layer_config.info.fb.addr[0] = (unsigned long long )SunxiMemGetPhysicAddressCpu(ops,sunxifbsink->rotate_addr_phy[0]);
		layer_config.info.fb.addr[1] = (unsigned long long )trans_info.dst_frame.laddr[1];
		layer_config.info.fb.addr[2] = (unsigned long long )trans_info.dst_frame.laddr[2];
	
		layer_config.info.fb.size[0].width = trans_info.dst_frame.pitch[0];
		layer_config.info.fb.size[1].width = trans_info.dst_frame.pitch[1];
		layer_config.info.fb.size[2].width = trans_info.dst_frame.pitch[2];
	
		layer_config.info.fb.size[0].height = trans_info.dst_frame.height[0];
		layer_config.info.fb.size[1].height = trans_info.dst_frame.height[1];
		layer_config.info.fb.size[2].height = trans_info.dst_frame.height[2];
	}
	//initialize layer info
	layer_config.info.mode = LAYER_MODE_BUFFER;
	layer_config.info.zorder = 1;
	layer_config.info.alpha_mode = 1;
	layer_config.info.alpha_value = 0xff;

	layer_config.info.fb.crop.x = 0;
	layer_config.info.fb.crop.y = 0;
	if(rotate_enable && (trans_info.mode == TR_ROT_90 || trans_info.mode == TR_ROT_270))
	{
		layer_config.info.fb.crop.width = (unsigned long long)rect_height << 32;
		layer_config.info.fb.crop.height = (unsigned long long)rect_width << 32;
	}
	else
	{
		layer_config.info.fb.crop.width = (unsigned long long)rect_width << 32;
		layer_config.info.fb.crop.height = (unsigned long long)rect_height << 32;
	}
	layer_config.info.fb.color_space = (framebuffersink->video_rectangle.h < 720) ? DISP_BT601 : DISP_BT709;
		
	layer_config.info.screen_win.x = framebuffersink->video_rectangle.x;
	layer_config.info.screen_win.y = framebuffersink->video_rectangle.y;
	layer_config.info.screen_win.width = framebuffersink->video_rectangle.w;
	layer_config.info.screen_win.height = framebuffersink->video_rectangle.h;

	layer_config.enable = TRUE;
	layer_config.layer_id = sunxifbsink->layer_id;
	layer_config.channel = sunxifbsink->framebuffer_id;
	
	layer_config.info.fb.flags= DISP_BF_NORMAL;
	layer_config.info.fb.scan= DISP_SCAN_PROGRESSIVE;
	
	tmp[0] = sunxifbsink->framebuffer_id;
	tmp[1] = (unsigned long)&layer_config;
	tmp[2] = 1;
	if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_SET_CONFIG, (void *)tmp) < 0)
	  return FALSE;

	gst_sunxifbsink_show_layer(sunxifbsink);
  	return GST_FLOW_OK;
}



static GstFlowReturn
gst_sunxifbsink_show_overlay_yuv_planar (GstFramebufferSink *framebuffersink,
    guintptr framebuffer_offset, GstVideoFormat format)
{
  GstSunxifbsink *sunxifbsink = GST_SUNXIFBSINK (framebuffersink);
  GstFbdevFramebufferSink *fbdevframebuffersink =
      GST_FBDEVFRAMEBUFFERSINK (framebuffersink);
  	struct disp_layer_config layer_config;
  	unsigned long tmp[4];
	/*gchar s[256];*/

	memset(&layer_config, 0, sizeof(struct disp_layer_config));

    if (format == GST_VIDEO_FORMAT_Y444) {
      layer_config.info.fb.addr[0] = framebuffer_offset;
      layer_config.info.fb.addr[1] = framebuffer_offset
          + framebuffersink->overlay_plane_offset[1];
      layer_config.info.fb.addr[2] = framebuffer_offset
          + framebuffersink->overlay_plane_offset[2];
	  layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
        / (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
        0, 8)
        * GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
      layer_config.info.fb.size[1].width = framebuffersink->overlay_scanline_stride[0]
	  	/ (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
		0, 8)
		* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
	  layer_config.info.fb.size[2].width = framebuffersink->overlay_scanline_stride[0]
        / (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
        0, 8)
        * GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
	  layer_config.info.fb.size[0].height = framebuffersink->videosink.height;
      layer_config.info.fb.size[1].height = framebuffersink->videosink.height;
	  layer_config.info.fb.size[2].height = framebuffersink->videosink.height;
      layer_config.info.fb.format = DISP_FORMAT_YUV444_P;
    }
    else if (format == GST_VIDEO_FORMAT_NV12
        || format == GST_VIDEO_FORMAT_NV21) {
      if(format == GST_VIDEO_FORMAT_NV12)  
      	layer_config.info.fb.format = DISP_FORMAT_YUV420_SP_UVUV;
	  else
	  	layer_config.info.fb.format = DISP_FORMAT_YUV420_SP_VUVU;
      layer_config.info.fb.addr[0] = framebuffer_offset;
      layer_config.info.fb.addr[1] = framebuffer_offset
          + framebuffersink->overlay_plane_offset[1];

	  layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
        / (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
        0, 8)
        * GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
	  layer_config.info.fb.size[1].width = (framebuffersink->overlay_scanline_stride[0]
        / (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
        0, 8)
        * GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8))/2;
	  
      layer_config.info.fb.size[1].height = framebuffersink->videosink.height/2;
	  layer_config.info.fb.size[2].height = framebuffersink->videosink.height/2;

	  /*g_sprintf(s,"-->yuv planar(fb_off=0x%08X),start=0x%08x,p1_of=%x,p2_of=%x,
	    str1=%d,vh=%d,w0=%d,(%d %d,%dx%d),%dx%d",framebuffer_offset, \
	  	fbdevframebuffersink->fixinfo.smem_start, \
	  	framebuffersink->overlay_plane_offset[1], \
	  	framebuffersink->overlay_plane_offset[2], \
	  	framebuffersink->overlay_scanline_stride[0], \
	  	framebuffersink->videosink.height, \
	  	layer_config.info.fb.size[0].width, \
	  	framebuffersink->video_rectangle.x, \
	  	framebuffersink->video_rectangle.y, \
	  	framebuffersink->video_rectangle.w, \
	  	framebuffersink->video_rectangle.h, \
	  	framebuffersink->videosink.width,framebuffersink->videosink.height);
	  GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, s);*/
	
    }
    else {
	  layer_config.info.fb.format = DISP_FORMAT_YUV420_P;	
      layer_config.info.fb.addr[0] = framebuffer_offset;
      if (format == GST_VIDEO_FORMAT_I420) {
        layer_config.info.fb.addr[1] = framebuffer_offset + framebuffersink->overlay_plane_offset[1];
        layer_config.info.fb.addr[2] = framebuffer_offset + framebuffersink->overlay_plane_offset[2];
		layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
        / (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
        0, 8)
        * GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
		layer_config.info.fb.size[0].height = framebuffersink->videosink.height;
	  	layer_config.info.fb.size[1].width = framebuffersink->videosink.width/2;
      	layer_config.info.fb.size[1].height = framebuffersink->videosink.height/2;
		layer_config.info.fb.size[2].width= framebuffersink->videosink.width/2;
	  	layer_config.info.fb.size[2].height = framebuffersink->videosink.height/2;
      }
      else {
        /* GST_VIDEO_FORMAT_YV12 */
        layer_config.info.fb.addr[1] = framebuffer_offset + framebuffersink->overlay_plane_offset[2];
        layer_config.info.fb.addr[2] = framebuffer_offset + framebuffersink->overlay_plane_offset[1];

		layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
        / (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo,
        0, 8)* GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
		layer_config.info.fb.size[0].height = framebuffersink->videosink.height;
	  	layer_config.info.fb.size[1].width = framebuffersink->videosink.width/2;
      	layer_config.info.fb.size[1].height = framebuffersink->videosink.height/2;
		layer_config.info.fb.size[2].width= framebuffersink->videosink.width/2;
	  	layer_config.info.fb.size[2].height = framebuffersink->videosink.height/2;
      }
    }
    //initialize layer info
 	layer_config.info.mode = LAYER_MODE_BUFFER;
	layer_config.info.zorder = 1;
	layer_config.info.alpha_mode = 1;
	layer_config.info.alpha_value = 0xff;

	layer_config.info.fb.crop.x = 0;
    layer_config.info.fb.crop.y = 0;
    layer_config.info.fb.crop.width = (unsigned long long)framebuffersink->videosink.width << 32;
    layer_config.info.fb.crop.height = (unsigned long long)framebuffersink->videosink.height << 32;
	layer_config.info.fb.color_space = (framebuffersink->video_rectangle.h < 720) ? DISP_BT601 : DISP_BT709;
		
    layer_config.info.screen_win.x = framebuffersink->video_rectangle.x;
    layer_config.info.screen_win.y = framebuffersink->video_rectangle.y;
    layer_config.info.screen_win.width = framebuffersink->video_rectangle.w;
    layer_config.info.screen_win.height = framebuffersink->video_rectangle.h;

	layer_config.enable = TRUE;
	layer_config.layer_id = sunxifbsink->layer_id;
  	layer_config.channel = sunxifbsink->framebuffer_id;
	
	layer_config.info.fb.flags= DISP_BF_NORMAL;
	layer_config.info.fb.scan= DISP_SCAN_PROGRESSIVE;
	
	tmp[0] = sunxifbsink->framebuffer_id;
	tmp[1] = (unsigned long)&layer_config;
	tmp[2] = 1;
	if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_SET_CONFIG, (void *)tmp) < 0)
	  return FALSE;

    gst_sunxifbsink_show_layer(sunxifbsink);

  return GST_FLOW_OK;
}

static GstFlowReturn
gst_sunxifbsink_show_overlay_yuv_packed (GstFramebufferSink *framebuffersink,
    guintptr framebuffer_offset, GstVideoFormat format)
{
  GstSunxifbsink *sunxifbsink = GST_SUNXIFBSINK (framebuffersink);
  GstFbdevFramebufferSink *fbdevframebuffersink =
      GST_FBDEVFRAMEBUFFERSINK (framebuffersink);
    struct disp_layer_config layer_config;
    unsigned long tmp[4];

	memset(&layer_config, 0, sizeof(struct disp_layer_config));

    layer_config.info.fb.addr[0] = framebuffer_offset;
    layer_config.info.fb.size[0].height = framebuffersink->videosink.height;
	layer_config.info.fb.size[0].width = framebuffersink->overlay_scanline_stride[0]
        / (GST_VIDEO_FORMAT_INFO_SCALE_WIDTH (framebuffersink->video_info.finfo, 0, 8)
        * GST_VIDEO_INFO_COMP_PSTRIDE (&framebuffersink->video_info, 0) / 8);
		
    if (format == GST_VIDEO_FORMAT_AYUV) {
      layer_config.info.fb.format = DISP_FORMAT_YUV444_P;
      GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "---->AYUV");
    }
    else {
      layer_config.info.fb.format = DISP_FORMAT_YUV422_P;
      if (format == GST_VIDEO_FORMAT_YUY2)
       GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "---->YUY2");
    }

    //initialize layer info
 	layer_config.info.mode = LAYER_MODE_BUFFER;
	layer_config.info.zorder = 1;
	layer_config.info.alpha_mode = 1;
	layer_config.info.alpha_value = 0xff;

	layer_config.info.fb.crop.x = 0;
    layer_config.info.fb.crop.y = 0;
    layer_config.info.fb.crop.width = (unsigned long long)framebuffersink->videosink.width << 32;
    layer_config.info.fb.crop.height = (unsigned long long)framebuffersink->videosink.height << 32;
	layer_config.info.fb.color_space = (framebuffersink->video_rectangle.h < 720) ? DISP_BT601 : DISP_BT709;
		
    layer_config.info.screen_win.x = framebuffersink->video_rectangle.x;
    layer_config.info.screen_win.y = framebuffersink->video_rectangle.y;
    layer_config.info.screen_win.width = framebuffersink->video_rectangle.w;
    layer_config.info.screen_win.height = framebuffersink->video_rectangle.h;

	layer_config.enable = TRUE;
	layer_config.layer_id = sunxifbsink->layer_id;
  	layer_config.channel = sunxifbsink->framebuffer_id;
	
	layer_config.info.fb.flags= DISP_BF_NORMAL;
	layer_config.info.fb.scan= DISP_SCAN_PROGRESSIVE;

	tmp[0] = sunxifbsink->framebuffer_id;
	tmp[1] = (unsigned long)&layer_config;
	tmp[2] = 1;
	if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_SET_CONFIG, (void *)tmp) < 0)
	  return FALSE;

    gst_sunxifbsink_show_layer(sunxifbsink);

  return GST_FLOW_OK;
}

static GstFlowReturn
gst_sunxifbsink_show_overlay_bgrx32 (GstFramebufferSink *framebuffersink,
    guintptr framebuffer_offset)
{
  GstSunxifbsink *sunxifbsink = GST_SUNXIFBSINK (framebuffersink);
  GstFbdevFramebufferSink *fbdevframebuffersink =
      GST_FBDEVFRAMEBUFFERSINK (framebuffersink);
  	struct disp_layer_config layer_config;
    unsigned long tmp[4];
	GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "-->sunxisink_show_overlay_bgrx32");

    memset(&layer_config, 0, sizeof(struct disp_layer_config));

    /* BGRX layer. */
	layer_config.info.mode = LAYER_MODE_BUFFER;
	layer_config.info.zorder = 1;
	layer_config.info.alpha_mode = 1;
	layer_config.info.alpha_value = 0xff;
	
    layer_config.info.fb.addr[0] = framebuffer_offset;
    layer_config.info.fb.size[sunxifbsink->framebuffer_id].width = framebuffersink->overlay_scanline_stride[0] >> 2;
    layer_config.info.fb.size[sunxifbsink->framebuffer_id].height = framebuffersink->videosink.height;
    layer_config.info.fb.format = DISP_FORMAT_ARGB_8888;

	layer_config.info.fb.crop.x = 0;
    layer_config.info.fb.crop.y = 0;
    layer_config.info.fb.crop.width = (unsigned long long)framebuffersink->videosink.width << 32;
    layer_config.info.fb.crop.height = (unsigned long long)framebuffersink->videosink.height << 32;
	layer_config.info.fb.color_space = (framebuffersink->video_rectangle.h < 720) ? DISP_BT601 : DISP_BT709;
		
    layer_config.info.screen_win.x = framebuffersink->video_rectangle.x;
    layer_config.info.screen_win.y = framebuffersink->video_rectangle.y;
    layer_config.info.screen_win.width = framebuffersink->video_rectangle.w;
    layer_config.info.screen_win.height = framebuffersink->video_rectangle.h;
	
	layer_config.enable = TRUE;
	layer_config.layer_id = sunxifbsink->layer_id;
  	layer_config.channel = sunxifbsink->framebuffer_id;
	layer_config.info.fb.flags= DISP_BF_NORMAL;
	layer_config.info.fb.scan= DISP_SCAN_PROGRESSIVE;

	tmp[0] = sunxifbsink->framebuffer_id;
	tmp[1] = (unsigned long)&layer_config;
	tmp[2] = 1;
	if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_SET_CONFIG, (void *)tmp) < 0)
	  return FALSE;

    gst_sunxifbsink_show_layer(sunxifbsink);

  return GST_FLOW_OK;
}

static GstFlowReturn
gst_sunxifbsink_show_overlay (GstFramebufferSink *framebuffersink,
    GstMemory *memory)
{
  GstSunxifbsink *sunxifbsink = GST_SUNXIFBSINK (framebuffersink);
  GstFbdevFramebufferSink *fbdevframebuffersink =
      GST_FBDEVFRAMEBUFFERSINK (framebuffersink);
  GstMapInfo mapinfo;
  guintptr framebuffer_offset;
  GstFlowReturn res;
  struct SunxiMemOpsS* ops =  GetMemAdapterOpsS();
  
  gst_memory_map(memory, &mapinfo, 0);
  framebuffer_offset = mapinfo.data - fbdevframebuffersink->framebuffer;

 GST_LOG_OBJECT (sunxifbsink, "Show overlay called (offset = 0x%08X)",
      framebuffer_offset);
  
  res = GST_FLOW_ERROR;
  if(GST_MEMORY_FLAG_IS_SET(memory, GST_MEMORY_FLAG_PHYSICALLY_CONTIGUOUS))
  {
		gst_sunxifbsink_show_memory_yuv_planar(framebuffersink,
			sunxifbsink->overlay_format,memory);
  }
  else
  {
	  if(framebuffersink->max_video_memory_property <= 0)
	  {
		framebuffer_offset+=fbdevframebuffersink->fixinfo.smem_start;
	  }
	  else
	  {
	    SunxiMemFlushCache(ops, mapinfo.data, framebuffersink->video_info.size);
		framebuffer_offset = SunxiMemGetPhysicAddressCpu(ops, mapinfo.data);
	  }
  
	  if (sunxifbsink->overlay_format == GST_VIDEO_FORMAT_I420 ||
	      sunxifbsink->overlay_format == GST_VIDEO_FORMAT_YV12 ||
	      sunxifbsink->overlay_format == GST_VIDEO_FORMAT_Y444 ||
	      sunxifbsink->overlay_format == GST_VIDEO_FORMAT_NV12 ||
	      sunxifbsink->overlay_format == GST_VIDEO_FORMAT_NV21)
	    res =  gst_sunxifbsink_show_overlay_yuv_planar (framebuffersink,
	        framebuffer_offset, sunxifbsink->overlay_format);
	  else if (sunxifbsink->overlay_format == GST_VIDEO_FORMAT_YUY2 ||
	      sunxifbsink->overlay_format == GST_VIDEO_FORMAT_UYVY ||
	      sunxifbsink->overlay_format == GST_VIDEO_FORMAT_AYUV)
	    res =  gst_sunxifbsink_show_overlay_yuv_packed (framebuffersink,
	        framebuffer_offset, sunxifbsink->overlay_format);
	  else if (sunxifbsink->overlay_format == GST_VIDEO_FORMAT_BGRx)
	    res = gst_sunxifbsink_show_overlay_bgrx32 (framebuffersink,
	        framebuffer_offset);
  }

  gst_memory_unmap(memory, &mapinfo);
  return res;
}

static gboolean
gst_sunxifbsink_reserve_layer(GstSunxifbsink *sunxifbsink) {
    struct disp_layer_config layer_config;
    unsigned long tmp[4];
	unsigned int screen_w, screen_h;
	gchar s[256];
	
	tmp[0]=0;
	screen_w = ioctl(sunxifbsink->fd_disp, DISP_GET_SCN_WIDTH, (void *)tmp);
	screen_h = ioctl(sunxifbsink->fd_disp, DISP_GET_SCN_HEIGHT, (void *)tmp);

	 g_sprintf(s,"reserver layer called (screen = %d x %d)",screen_w, screen_h);
	 GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, s);
	 
    /* try to allocate a layer */
    layer_config.info.screen_win.x = 0;
    layer_config.info.screen_win.y = 0;
    layer_config.info.screen_win.width = screen_w;
    layer_config.info.screen_win.height = screen_h;

	layer_config.info.fb.crop.x = 0;
    layer_config.info.fb.crop.y = 0;
    layer_config.info.fb.crop.width = (unsigned long long)screen_w << 32;
    layer_config.info.fb.crop.height = (unsigned long long)screen_h << 32;
	
	layer_config.enable = TRUE;
	layer_config.layer_id = sunxifbsink->layer_id;
  	layer_config.channel = sunxifbsink->framebuffer_id;
	layer_config.info.mode= LAYER_MODE_BUFFER;
	layer_config.info.fb.flags= DISP_BF_NORMAL;
	layer_config.info.fb.scan= DISP_SCAN_PROGRESSIVE;
	layer_config.info.fb.format = DISP_FORMAT_ARGB_8888;
	layer_config.info.zorder = 1;
	layer_config.info.alpha_mode = 1;
	layer_config.info.alpha_value = 0xff;

	tmp[0] = sunxifbsink->framebuffer_id;
  	tmp[1] = (unsigned long)&layer_config;
  	tmp[2] = 1;
  	if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_SET_CONFIG, (void *)tmp) < 0)
    	return FALSE;
	
    sunxifbsink->layer_has_scaler = TRUE;
  return TRUE;
}

static void
gst_sunxifbsink_release_layer(GstSunxifbsink *sunxifbsink) {
    unsigned long tmp[4];
	GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "-->sunxifbsink_release_layer");	
    tmp[0] = sunxifbsink->framebuffer_id;
    tmp[1] = sunxifbsink->layer_id;
    ioctl (sunxifbsink->fd_disp, DISP_LAYER_DISABLE, tmp);

    sunxifbsink->layer_id = -1;
    sunxifbsink->layer_has_scaler = 0;
}

static gboolean gst_sunxifbsink_show_layer(GstSunxifbsink *sunxifbsink) {
  unsigned long tmp[4];
  gchar s[256];
  
  struct disp_layer_config sunxidsp_layer_config;
  	  
  if (sunxifbsink->layer_is_visible)
    return TRUE;

  if (sunxifbsink->layer_id < 0)
    return FALSE;

  g_sprintf(s,"-->show_layer id %d)",sunxifbsink->layer_id);
  GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, s);

  memset(&sunxidsp_layer_config, 0, sizeof(struct disp_layer_config));
  
  sunxidsp_layer_config.channel =0;
  sunxidsp_layer_config.layer_id = 0;
  tmp[0] = sunxifbsink->framebuffer_id;
  tmp[1] = (unsigned long)&sunxidsp_layer_config;
  tmp[2] = 1;
  tmp[3] = 0;
  if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_GET_CONFIG, &tmp) < 0)
    return FALSE;
  
  sunxidsp_layer_config.enable = TRUE;
  sunxidsp_layer_config.layer_id = sunxifbsink->layer_id;
  sunxidsp_layer_config.channel = sunxifbsink->framebuffer_id;

  tmp[0] = sunxifbsink->framebuffer_id;
  tmp[1] = (unsigned long)&sunxidsp_layer_config;
  tmp[2] = 1;
  if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_SET_CONFIG, (void *)tmp) < 0)
    return FALSE;

  sunxifbsink->layer_is_visible = TRUE;
  return TRUE;
}

static void gst_sunxifbsink_hide_layer(GstSunxifbsink *sunxifbsink) {
  unsigned long tmp[4];
  struct disp_layer_config sunxidsp_layer_config;
  GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "-->sunxifbsink_hide_layer");   

  if (!sunxifbsink->layer_is_visible)
    return;

  memset(&sunxidsp_layer_config, 0, sizeof(struct disp_layer_config));

  sunxidsp_layer_config.channel =0;
  sunxidsp_layer_config.layer_id = 0;
  tmp[0] = sunxifbsink->framebuffer_id;
  tmp[1] = (unsigned long)&sunxidsp_layer_config;
  tmp[2] = 1;
  tmp[3] = 0;

  if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_GET_CONFIG, &tmp) < 0) {
  	GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "Get config failed");
    return;
  	}
  sunxidsp_layer_config.enable = FALSE;
  sunxidsp_layer_config.layer_id = sunxifbsink->layer_id;
  sunxidsp_layer_config.channel = sunxifbsink->framebuffer_id;

  tmp[0] = sunxifbsink->framebuffer_id;
  tmp[1] = (unsigned long)&sunxidsp_layer_config;
  tmp[2] = 1;
  if (ioctl(sunxifbsink->fd_disp, DISP_LAYER_SET_CONFIG, (void *)tmp) < 0){
  	GST_SUNXIFBSINK_MESSAGE_OBJECT (sunxifbsink, "Set config failed");
    return;
  	}
  sunxifbsink->layer_is_visible = FALSE;
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  /* Remember to set the rank if it's an element that is meant
     to be autoplugged by decodebin. */
  return gst_element_register (plugin, "sunxifbsink", GST_RANK_SECONDARY,
      GST_TYPE_SUNXIFBSINK);
}

/* these are normally defined by the GStreamer build system.
   If you are creating an element to be included in gst-plugins-*,
   remove these, as they're always defined.  Otherwise, edit as
   appropriate for your external plugin package. */
#ifndef VERSION
#define VERSION "0.1"
#endif
#ifndef PACKAGE
#define PACKAGE "gstsunxifbsink"
#endif
#ifndef PACKAGE_NAME
#define PACKAGE_NAME "gstreamer1.0-sunxi-plugins"
#endif
#ifndef GST_PACKAGE_ORIGIN
#define GST_PACKAGE_ORIGIN "https://github.com/hglm"
#endif

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    sunxifbsink,
    "Accelerated console framebuffer video sink for sunxi-based devices",
    plugin_init, VERSION, "LGPL", PACKAGE_NAME, GST_PACKAGE_ORIGIN)

